/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.techfak.scie.classifiers;

import de.unibi.techfak.scie.classifiers.data.ClassificationResult;
import de.unibi.techfak.scie.classifiers.data.LabeledDataPoint;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielfeld.de
 */
public class ClassifierEvaluation {

	public static final String TRAIN_SEPARATOR = "--- Training ---";
	public static final String TEST_SEPARATOR = "--- Test ---";
	public static final int DEFAULT_ROC_STEPS = 51;

	private final SetEvaluation trainEvaluation;
	private final SetEvaluation testEvaluation;

	public ClassifierEvaluation(Classifier classifier,
			ArrayList<LabeledDataPoint> trainingData,
			ArrayList<LabeledDataPoint> testData) {
		this(classifier, trainingData, testData, DEFAULT_ROC_STEPS, false);
	}

	public ClassifierEvaluation(Classifier classifier,
			ArrayList<LabeledDataPoint> trainingData,
			ArrayList<LabeledDataPoint> testData,
			boolean verbose) {
		this(classifier, trainingData, testData, DEFAULT_ROC_STEPS, verbose);
	}

	public ClassifierEvaluation(Classifier classifier,
			ArrayList<LabeledDataPoint> trainingData,
			ArrayList<LabeledDataPoint> testData,
			int roc_steps) {
		this(classifier, trainingData, testData, roc_steps, false);
	}

	public ClassifierEvaluation(Classifier classifier,
			ArrayList<LabeledDataPoint> trainingData,
			ArrayList<LabeledDataPoint> testData,
			int roc_steps, boolean verbose) {
		this.trainEvaluation = new SetEvaluation(classifier, trainingData, roc_steps, verbose);
		this.testEvaluation = new SetEvaluation(classifier, testData, roc_steps, verbose);
	}

	public ClassifierEvaluation(Reader in) throws IOException {
		final BufferedReader bufIn;
		if (in instanceof BufferedReader) {
			bufIn = (BufferedReader) in;
		} else {
			bufIn = new BufferedReader(in);
		}
		final StringBuilder buffer = new StringBuilder();
		String line;
		boolean start = false;
		while ((line = bufIn.readLine()) != null) {
			if (line.equals(TRAIN_SEPARATOR)) {
				start = true;
				continue;
			}
			if (line.equals(TEST_SEPARATOR)) {
				break;
			}
			if (start) {
				buffer.append(line);
			}
		}
		final BufferedReader trainIn = new BufferedReader(
				new StringReader(buffer.toString()));
		this.trainEvaluation = new SetEvaluation(trainIn);
		this.testEvaluation = new SetEvaluation(bufIn);
	}

	public SetEvaluation getTestEvaluation() {
		return testEvaluation;
	}

	public SetEvaluation getTrainEvaluation() {
		return trainEvaluation;
	}

	@Override
	public String toString() {
		return "Training Results:\n" + trainEvaluation.toString()
				+ "\nTest Results:\n" + testEvaluation.toString();
	}

	public void store(Writer out) throws IOException {
		out.write(TRAIN_SEPARATOR);
		out.write("\n");
		trainEvaluation.store(out);
		out.write("\n");
		out.write(TEST_SEPARATOR);
		out.write("\n");
		testEvaluation.store(out);
	}

	/**
	 * This object encapsulates evaluation measures for one set, either training
	 * or test. Namely:
	 * <ul><li>number of data points (N)</li>
	 * <li>accuracy</li>
	 * <li>precision</li>
	 * <li>recall</li>
	 * <li>F1-measure</li>
	 * <li>reciever operator curve data:
	 * <ul><li>x axis (recall)</li>
	 * <li>y axis (specificity)</li></ul>
	 * </li></ul>
	 */
	public static class SetEvaluation {

		public static final String N_STR = "N";
		private final int N;
		public static final String PRECISION_STR = "Precision";
		private final double precision;
		public static final String RECALL_STR = "Recall";
		private final double recall;
		public static final String F1_STR = "F1";
		public static final String ACC_STR = "Accuracy";
		private final double accuracy;
		public static final String ROC_THR_STR = "ROC Confidence Thresholds";
		private final double[] roc_thresholds;
		public static final String ROC_SPEC_STR = "ROC Specificities";
		private final double[] roc_specificities;
		public static final String ROC_REC_STR = "ROC Recalls";
		private final double[] roc_recalls;
		public static final String AREA_ROC_STR = "Area under ROC";
		private final double area_under_roc;

		public SetEvaluation(Classifier classifier,
				ArrayList<LabeledDataPoint> data,
				int roc_steps) {
			this(classifier, data, roc_steps, false);
		}

		public SetEvaluation(Classifier classifier,
				ArrayList<LabeledDataPoint> data,
				int roc_steps, boolean verbose) {
			this.N = data.size();
			/*
			 * classify all datapoints once and build a list of tuples of data
			 * point and classification.
			 */
			final ArrayList<Tuple<LabeledDataPoint, ClassificationResult>> classifications = new ArrayList<>(N);

			if (classifier instanceof LibLinearClassifier) {
				((LibLinearClassifier) classifier).setSilentCorrect(false);
			}
			for (final LabeledDataPoint test_p : data) {
				final ClassificationResult classification = classifier.classify(
						test_p.getData());
				if (!verbose && classifier instanceof LibLinearClassifier) {
					((LibLinearClassifier) classifier).setSilentCorrect(true);
				}
				classifications.add(new Tuple<>(test_p, classification));
			}
			//sort according to confidence.
			Collections.sort(classifications, new ConfidenceComparator());
			/*
			 * evaluate true positives, true negatives, false negatives and
			 * false positives
			 */
			double tp = 0;
			double tn = 0;
			double fn = 0;
			double fp = 0;
			if (verbose) {
				System.out.println("TrueLabel;ClassifierLabel;Confidence");
			}
			for (final Tuple<LabeledDataPoint, ClassificationResult> classification : classifications) {
				if (verbose) {
					System.out.println(classification.getFirst().getLabel() + ";"
							+ classification.getSecond().getLabel() + ";"
							+ classification.getSecond().getConfidence());
				}
				if (classification.getFirst().getLabel()) {
					if (classification.getSecond().getLabel()) {
						tp++;
					} else {
						fn++;
					}
				} else {
					if (classification.getSecond().getLabel()) {
						fp++;
					} else {
						tn++;
					}
				}
			}
			if (tp > 0) {
				this.precision = tp / (tp + fp);
				this.recall = tp / (tp + fn);
			} else {
				this.precision = 0;
				this.recall = 0;
			}
			this.accuracy = (double) (tp + tn) / (double) N;
			//calculate reciever operator curve
			final double roc_step_size = 1 / (double) (roc_steps - 1);
			this.roc_thresholds = new double[roc_steps];
			this.roc_recalls = new double[roc_steps];
			this.roc_specificities = new double[roc_steps];
			roc_thresholds[0] = 1;
			tp = 0;
			tn = 0;
			fn = 0;
			fp = 0;
			int last_threshold_idx = 0;
			/*
			 * initialize true negatives and false negatives (because initially,
			 * we treat everything as a negative; confidence threshold is 1)
			 */
			for (final LabeledDataPoint point : data) {
				if (point.getLabel()) {
					fn++;
				} else {
					tn++;
				}
			}
			//Thus the initial specificity is 1 and the initial recall is zero.
			roc_specificities[0] = 1;
			/*
			 * At the end the confidence threshold is 0, so the recall is 1 and
			 * the specificity is 0.
			 */
			roc_recalls[roc_steps - 1] = 1;
			/*
			 * Now iterate through all intermediate levels. Treat all datapoints
			 * above the currently tested confidence threshold as positive
			 * classifications and all below as negative classifications.
			 *
			 * As the number of negatives monotonously decreases and -
			 * accordingly - the number of positives increases we can just count
			 * the number of positives that we got by lowering the threshold
			 * in the current step.
			 */
			for (int r = 1; r < roc_steps - 1; r++) {
				roc_thresholds[r] = roc_thresholds[r - 1] - roc_step_size;
				int threshold_idx = last_threshold_idx;
				double confidence = 0;
				while (threshold_idx < N) {
					final Tuple<LabeledDataPoint, ClassificationResult> classification
							= classifications.get(threshold_idx);
					confidence = classification.getSecond().getConfidence();
					if (confidence < roc_thresholds[r]) {
						break;
					}
					if (classification.getFirst().getLabel()) {
						/*
						 * if this is a positive, then we have correctly
						 * classified it as a positive now.
						 */
						tp++;
						fn--;
					} else {
						/*
						 * If this is a negative, we have wrongly classified
						 * it as a positive now.
						 */
						tn--;
						fp++;
					}
					threshold_idx++;
				}
				//if we have found the threshold, compute current recall and specificity.
				if (tp > 0) {
					roc_recalls[r] = (double) tp / (double) (tp + fn);
				}
				if (tn > 0) {
					roc_specificities[r] = (double) tn / (double) (tn + fp);
				}
				last_threshold_idx = threshold_idx;
			}
			/*
			 * compute area under curve using an approximation by rectangles.
			 * The x-axis of an Reviecer Operator Curve is the false positive
			 * rate (1 minus specificity) and the y axis is the recall.
			 * Thus our rectangles have the width
			 *
			 * (1 - roc_specificties[r]) - (1 - roc_specificties[r-1])
			 * = roc_specificties[r-1] - roc_specificties[r]
			 *
			 * and the height
			 *
			 * average(roc_recalls[r-1],roc_recalls[r])
			 */
			double area_under_roc = 0;
			for (int r = 1; r < roc_steps; r++) {
				//the widt
				final double width = roc_specificities[r - 1] - roc_specificities[r];
				final double height = 0.5 * (roc_recalls[r - 1] + roc_recalls[r]);
				area_under_roc += width * height;
			}
			this.area_under_roc = area_under_roc;
		}

		public SetEvaluation(int N,
				double precision, double recall, double accuracy,
				double[] roc_thresholds, double[] roc_specificities,
				double[] roc_recalls, double area_under_roc) {
			this.N = N;
			this.precision = precision;
			this.recall = recall;
			this.accuracy = accuracy;
			this.roc_thresholds = roc_thresholds;
			this.roc_specificities = roc_specificities;
			this.roc_recalls = roc_recalls;
			this.area_under_roc = area_under_roc;
		}

		public SetEvaluation(Reader in) throws IOException {
			final BufferedReader bufIn;
			if (in instanceof BufferedReader) {
				bufIn = (BufferedReader) in;
			} else {
				bufIn = new BufferedReader(in);
			}
			final HashMap<String, String> contentMap = new HashMap<>();
			String line;
			while ((line = bufIn.readLine()) != null) {
				final String[] split = line.split(" = ");
				if (split.length != 2) {
					continue;
				}
				contentMap.put(split[0], split[1]);
			}
			this.N = retrieveInt(contentMap, N_STR);
			this.precision = retrieveDouble(contentMap, PRECISION_STR);
			this.recall = retrieveDouble(contentMap, RECALL_STR);
			this.accuracy = retrieveDouble(contentMap, ACC_STR);
			this.roc_thresholds = retrieveDoubleArr(contentMap, ROC_THR_STR);
			this.roc_specificities = retrieveDoubleArr(contentMap, ROC_SPEC_STR);
			this.roc_recalls = retrieveDoubleArr(contentMap, ROC_REC_STR);
			if (roc_thresholds.length != roc_specificities.length) {
				throw new UnsupportedOperationException(
						"The input data was inconsistent: The number of "
						+ "confidence thresholds and the number of "
						+ "specificities did not match!");
			}
			if (roc_thresholds.length != roc_recalls.length) {
				throw new UnsupportedOperationException(
						"The input data was inconsistent: The number of "
						+ "confidence thresholds and the number of "
						+ "recalls did not match!");
			}
			this.area_under_roc = retrieveDouble(contentMap, AREA_ROC_STR);
		}

		private int retrieveInt(HashMap<String, String> contentMap, String key) {
			final String value = contentMap.get(key);
			if (value == null) {
				throw new UnsupportedOperationException(key
						+ " could not be retrieved from input!");
			}
			try {
				return Integer.parseInt(value);
			} catch (NumberFormatException ex) {
				throw new UnsupportedOperationException(key
						+ " could not be retrieved from input!", ex);
			}
		}

		private double retrieveDouble(HashMap<String, String> contentMap, String key) {
			final String value = contentMap.get(key);
			if (value == null) {
				throw new UnsupportedOperationException(key
						+ " could not be retrieved from input!");
			}
			try {
				return Double.parseDouble(value);
			} catch (NumberFormatException ex) {
				throw new UnsupportedOperationException(key
						+ " could not be retrieved from input!", ex);
			}
		}

		private double[] retrieveDoubleArr(HashMap<String, String> contentMap, String key) {
			final String value = contentMap.get(key);
			if (value == null) {
				throw new UnsupportedOperationException(key
						+ " could not be retrieved from input!");
			}
			try {
				final String[] split = value.replaceAll("[\\[\\]]", "").split(",");
				final double[] out = new double[split.length];
				for (int i = 0; i < split.length; i++) {
					out[i] = Double.parseDouble(split[i]);
				}
				return out;
			} catch (NumberFormatException ex) {
				throw new UnsupportedOperationException(key
						+ " could not be retrieved from input!", ex);
			}
		}

		public int getN() {
			return N;
		}

		public double getAccuracy() {
			return accuracy;
		}

		public double getPrecision() {
			return precision;
		}

		public double getRecall() {
			return recall;
		}

		public double getF1() {
			if (precision == 0 && recall == 0) {
				return 0;
			}
			return 2 * precision * recall / (precision + recall);
		}

		public double[] getRoc_thresholds() {
			return roc_thresholds;
		}

		public double[] getRoc_specificities() {
			return roc_specificities;
		}

		public double[] getRoc_recalls() {
			return roc_recalls;
		}

		public double getArea_under_roc() {
			return area_under_roc;
		}

		@Override
		public String toString() {
			return N_STR + " = " + N + "\n"
					+ ACC_STR + " = " + accuracy + "\n"
					+ PRECISION_STR + " = " + precision + "\n"
					+ RECALL_STR + " = " + recall + "\n"
					+ F1_STR + " = " + getF1() + "\n"
					+ AREA_ROC_STR + " = " + area_under_roc;
		}

		public void store(Writer out) throws IOException {
			out.write(N_STR + " = " + N + "\n");
			out.write(ACC_STR + " = " + accuracy + "\n");
			out.write(PRECISION_STR + " = " + precision + "\n");
			out.write(RECALL_STR + " = " + recall + "\n");
			out.write(F1_STR + " = " + getF1() + "\n");
			out.write(ROC_THR_STR + " = " + Arrays.toString(roc_thresholds) + "\n");
			out.write(ROC_SPEC_STR + " = " + Arrays.toString(roc_specificities) + "\n");
			out.write(ROC_REC_STR + " = " + Arrays.toString(roc_recalls) + "\n");
			out.write(AREA_ROC_STR + " = " + area_under_roc);
		}

		@Override
		public int hashCode() {
			int hash = 5;
			hash = 31 * hash + this.N;
			hash = 31 * hash + (int) (Double.doubleToLongBits(this.precision) ^ (Double.doubleToLongBits(this.precision) >>> 32));
			hash = 31 * hash + (int) (Double.doubleToLongBits(this.recall) ^ (Double.doubleToLongBits(this.recall) >>> 32));
			hash = 31 * hash + (int) (Double.doubleToLongBits(this.accuracy) ^ (Double.doubleToLongBits(this.accuracy) >>> 32));
			hash = 31 * hash + Arrays.hashCode(this.roc_thresholds);
			hash = 31 * hash + Arrays.hashCode(this.roc_specificities);
			hash = 31 * hash + Arrays.hashCode(this.roc_recalls);
			hash = 31 * hash + (int) (Double.doubleToLongBits(this.area_under_roc) ^ (Double.doubleToLongBits(this.area_under_roc) >>> 32));
			return hash;
		}

		@Override
		public boolean equals(Object obj) {
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			final SetEvaluation other = (SetEvaluation) obj;
			if (this.N != other.N) {
				return false;
			}
			if (Double.doubleToLongBits(this.precision) != Double.doubleToLongBits(other.precision)) {
				return false;
			}
			if (Double.doubleToLongBits(this.recall) != Double.doubleToLongBits(other.recall)) {
				return false;
			}
			if (Double.doubleToLongBits(this.accuracy) != Double.doubleToLongBits(other.accuracy)) {
				return false;
			}
			if (!Arrays.equals(this.roc_thresholds, other.roc_thresholds)) {
				return false;
			}
			if (!Arrays.equals(this.roc_specificities, other.roc_specificities)) {
				return false;
			}
			if (!Arrays.equals(this.roc_recalls, other.roc_recalls)) {
				return false;
			}
			if (Double.doubleToLongBits(this.area_under_roc) != Double.doubleToLongBits(other.area_under_roc)) {
				return false;
			}
			return true;
		}

	}

	private static class ConfidenceComparator implements Comparator<Tuple<LabeledDataPoint, ClassificationResult>> {

		@Override
		public int compare(Tuple<LabeledDataPoint, ClassificationResult> o1,
				Tuple<LabeledDataPoint, ClassificationResult> o2) {
			return -Double.compare(o1.getSecond().getConfidence(),
					o2.getSecond().getConfidence());
		}

	}

}
