/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.classifiers.data;

import de.unibi.techfak.scie.classifiers.FeatureUtils;
import de.unibi.techfak.scie.descriptors.Annotation;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Objects;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

/**
 * This models as a combination of core and slot for some relation. It is the
 * data interface for all CoreSlotClassifiers.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 * @param <C> core annotation class.
 * @param <S> slot annotation class.
 */
public abstract class CoreSlotCombinationDataPoint<C extends Annotation, S extends Annotation> implements DataPoint {

	private final C core;
	private final S slot;
	private final JCas jcas;
	private final double sigma;

	public CoreSlotCombinationDataPoint(C core, S slot, JCas jcas, double sigma) {
		this.core = core;
		this.slot = slot;
		this.jcas = jcas;
		this.sigma = sigma;
	}

	public C getCore() {
		return core;
	}

	public JCas getJcas() {
		return jcas;
	}

	public S getSlot() {
		return slot;
	}

	/**
	 * Returns a FeatureMap representing this CoreSlotCombination sparsely. This
	 * adds some standard features to the mix:
	 *
	 * <ul><li>Relative Start index in the text (for core and slot)</li>
	 * <li>Relative End index in the text (for core and slot)</li>
	 * <li>Relative annotation length (for core and slot)</li>
	 * <li>dictionary confidence (for core and slot); if these are not
	 * dictionary-based words, a 1 is used.</li>
	 * <li>character distance between core and slot.</li>
	 * <li>syntactics based distance between core and slot.</li>
	 * <li>words between core and slot (boolean features)</li>
	 * <li>core and slot word itself (boolean features)</li>
	 * <li>POS tags of core and slot word (boolean features)</li>
	 * <li>dictionary id identity (boolean features); if these are not
	 * dictionary-based words, a 0 is used.</li></ul>
	 *
	 * Additional features are added by the addSpecialFeatures method depending
	 * on the specific implementation.
	 *
	 * @return a FeatureMap representing this CoreSlotCombination sparsely.
	 */
	@Override
	public FeatureMap getFeatureRepresentation(FeatureDictionary dict,
			boolean addFeaturesToDictionary) {
		final FeatureMap features = new FeatureMap(dict);
		//add relative text position features.
		features.addNumericFeature("coreBegin",
				FeatureUtils.getRelativeBegin(core, jcas),
				addFeaturesToDictionary);
		features.addNumericFeature("coreEnd",
				FeatureUtils.getRelativeEnd(core, jcas),
				addFeaturesToDictionary);
		features.addNumericFeature("slotBegin",
				FeatureUtils.getRelativeBegin(slot, jcas),
				addFeaturesToDictionary);
		features.addNumericFeature("slotEnd",
				FeatureUtils.getRelativeEnd(core, jcas),
				addFeaturesToDictionary);
		//add length features
		features.addNumericFeature("coreLength",
				FeatureUtils.getRelativeLength(core, jcas),
				addFeaturesToDictionary);
		features.addNumericFeature("slotLength",
				FeatureUtils.getRelativeLength(slot, jcas),
				addFeaturesToDictionary);
		//get confidence features
		features.addNumericFeature("coreConfidence",
				FeatureUtils.getConfidence(core),
				addFeaturesToDictionary);
		features.addNumericFeature("slotConfidence",
				FeatureUtils.getConfidence(slot),
				addFeaturesToDictionary);
		//distance features
		features.addNumericFeature("charDist",
				FeatureUtils.getRelativeDist(core, slot, jcas),
				addFeaturesToDictionary);
		features.addNumericFeature("syntacticProbability",
				FeatureUtils.calculateSyntacticProbability(core, slot, sigma, jcas),
				addFeaturesToDictionary);

		//words in between
		for (final String w : FeatureUtils.getWordsInBetween(slot, core, jcas)) {
			features.addBooleanFeature("wordInBetween"
					+ FeatureUtils.cleanWord(w),
					addFeaturesToDictionary);
		}

		//get the words themselves.
		features.addBooleanFeature("core"
				+ FeatureUtils.cleanWord(core.getCoveredText()),
				addFeaturesToDictionary);
		features.addBooleanFeature("slot"
				+ FeatureUtils.cleanWord(slot.getCoveredText()),
				addFeaturesToDictionary);
		//get the POS tags of core and slot.
		for (final String pos : FeatureUtils.getPOSTags(core, jcas)) {
			features.addBooleanFeature("corePOS" + pos, addFeaturesToDictionary);
		}
		for (final String pos : FeatureUtils.getPOSTags(slot, jcas)) {
			features.addBooleanFeature("slotPOS" + pos, addFeaturesToDictionary);
		}
		//get dictionary id identity features
		features.addBooleanFeature("coreID" + FeatureUtils.getDictionaryID(core),
				addFeaturesToDictionary);
		features.addBooleanFeature("slotID" + FeatureUtils.getDictionaryID(slot),
				addFeaturesToDictionary);
		//add special features.
		addSpecialFeatures(features, addFeaturesToDictionary);
		return features;
	}

	/**
	 * This is to be implemented by a subclass to add domain-specific features
	 * to the mix.
	 *
	 * @param stdFeatures featureMap already filled with standard features.
	 * @param addFeaturesToDictionary this should be set to true during
	 * training, because then unknown features should be added to the
	 * FeatureDictionary. Outside training this should be set to false because
	 * then we want to treat every unknown feature as zero implicitly.
	 */
	public abstract void addSpecialFeatures(FeatureMap stdFeatures,
			boolean addFeaturesToDictionary);

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 97 * hash + Objects.hashCode(this.core);
		hash = 97 * hash + Objects.hashCode(this.slot);
		return hash;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final CoreSlotCombinationDataPoint<?, ?> other = (CoreSlotCombinationDataPoint<?, ?>) obj;
		if (!Objects.equals(this.core, other.core)) {
			return false;
		}
		if (!Objects.equals(this.slot, other.slot)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return "{core = " + core.getCoveredText() + " , slot = " + slot.
				getCoveredText() + "}";
	}

	/**
	 * The CoreSlotTrainingDataReader has a quadratic search space in nature: It
	 * combines all possible core instances found in the training data with all
	 * possible slot instances. For positive instances found in the given
	 * RawRelations this is not a problem, but the rest of the search space
	 * could in principle be treated as negative examples. To restrict the size
	 * of the search space to somewhat interesting datapoints we limit it to a
	 * maximum character distance between core and slot distance of
	 * MAXCHARDISTANCE.
	 *
	 * @param <C> the core annotation class.
	 * @param <S> the slot annotation class.
	 */
	public static abstract class CoreSlotTrainingDataReader<C extends Annotation, S extends Annotation>
			implements TrainingDataReader {

		public static final int MAXCHARDISTANCE = 300;

		private final Class<C> coreClass;
		private final Class<S> slotClass;
		private final String rawRelationClass;

		public CoreSlotTrainingDataReader(Class<C> coreClass, Class<S> slotClass, String rawRelationClass) {
			this.coreClass = coreClass;
			this.slotClass = slotClass;
			this.rawRelationClass = rawRelationClass;
		}

		public Class<C> getCoreClass() {
			return coreClass;
		}

		public String getRawRelationClass() {
			return rawRelationClass;
		}

		public Class<S> getSlotClass() {
			return slotClass;
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public void readTrainingData(JCas jcas,
				Collection<RawRelation> rawRelationInstances,
				Collection<LabeledDataPoint> trainingDataCollection) {
			final Collection<C> coreInstances = JCasUtil.select(jcas, coreClass);
			final HashMap<Long, C> coreIDMapping = new HashMap<>();
			for (final C coreInstance : coreInstances) {
				coreIDMapping.put(coreInstance.getAnnotationId(), coreInstance);
			}
			final Collection<S> slotInstances = JCasUtil.select(jcas, slotClass);
			final HashMap<Long, S> slotIDMapping = new HashMap<>();
			for (final S slotInstance : slotInstances) {
				slotIDMapping.put(slotInstance.getAnnotationId(), slotInstance);
			}
			final HashMap<Long, HashSet<Long>> positiveData = new HashMap<>();
			/*
			 * construct positive training examples.
			 */
			for (final RawRelation relationInstance : rawRelationInstances) {
				if (!relationInstance.relationClass.equals(rawRelationClass)) {
					continue;
				}
				/*
				 * Right now each raw relation might contain multiple candidates
				 * for core and slot. Ideally this can later be replaced
				 */
				final ArrayList<C> coreCandidates = new ArrayList<>();
				for (final Long id : relationInstance.ids) {
					final C coreCandidate = coreIDMapping.get(id);
					if (coreCandidate != null) {
						coreCandidates.add(coreCandidate);
					}
				}
				if (coreCandidates.isEmpty()) {
					continue;
				}
				final ArrayList<S> slotCandidates = new ArrayList<>();
				for (final Long id : relationInstance.ids) {
					final S slotCandidate = slotIDMapping.get(id);
					if (slotCandidate != null) {
						slotCandidates.add(slotCandidate);
					}
				}
				if (slotCandidates.isEmpty()) {
					continue;
				}
				for (final C coreCandidate : coreCandidates) {
					HashSet<Long> positiveSlotIDs = positiveData.get(coreCandidate.getAnnotationId());
					if (positiveSlotIDs == null) {
						positiveSlotIDs = new HashSet<>();
						positiveData.put(coreCandidate.getAnnotationId(), positiveSlotIDs);
					}
					for (final S slotCandidate : slotCandidates) {
						final CoreSlotCombinationDataPoint<C, S> combination
								= createCombination(coreCandidate, slotCandidate,
										jcas);
						positiveSlotIDs.add(slotCandidate.getAnnotationId());
						trainingDataCollection.add(
								new LabeledDataPoint(true, combination));
					}
				}
			}
			//construct negative training examples.
			for (final C coreInstance : coreInstances) {
				HashSet<Long> positiveSlotIDs
						= positiveData.get(coreInstance.getAnnotationId());
				if (positiveSlotIDs == null) {
					positiveSlotIDs = new HashSet<>();
				}
				for (final S slotInstance : slotInstances) {
					if (positiveSlotIDs.contains(slotInstance.getAnnotationId())) {
						continue;
					}
					//limit the number of negative data points by a rough distance filter.
					if (Math.abs(coreInstance.getBegin() - slotInstance.getBegin()) > MAXCHARDISTANCE) {
						continue;
					}
					final CoreSlotCombinationDataPoint<C, S> combination
							= createCombination(coreInstance, slotInstance,
									jcas);
					trainingDataCollection.add(
							new LabeledDataPoint(false, combination));
				}
			}
		}

		public abstract CoreSlotCombinationDataPoint<C, S> createCombination(
				C core, S slot, JCas jcas);

	}

}
