/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.classifiers.data;

import java.util.Collection;
import org.apache.uima.jcas.JCas;

/**
 * This is an interface for objects that are able to construct LabeledDataPoints
 * from manually annotated training data files.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielfeld.de
 */
public interface TrainingDataReader {

	/**
	 * This constructs training data for the given combination of XCAS document
	 * and RawRelations and writes them to the given collection of
	 * LabeledDataPoints. Note that the rawRelationInstances are treated as
	 * positive examples.
	 *
	 * Any additional information that might be required for constructing
	 * training data has to be set in the constructor of the implementing class.
	 *
	 * Also note that the space of possible negative data points might indeed be
	 * huge. Thus implementing classes might choose to restrict the number of
	 * negative data points by heuristics as maximum distance.
	 *
	 * The contract of this interface, however, garantuees that no positive
	 * examples specified in the rawRelationInstances are also constructed as
	 * negative training data.
	 *
	 * @param jcas an XCAS document given as JCas.
	 * @param rawRelationInstances a Collection of RawRelations that are treated
	 * as positive Training examples.
	 * @param trainingDataCollection
	 */
	public void readTrainingData(JCas jcas,
		Collection<RawRelation> rawRelationInstances,
		Collection<LabeledDataPoint> trainingDataCollection);
}
