/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.techfak.scie.classifiers;

import de.bwaldvogel.liblinear.SolverType;
import de.unibi.techfak.scie.descriptors.Annotation;
import de.unibi.techfak.scie.descriptors.Sentence;
import de.unibi.techfak.scie.descriptors.Token;
import de.unibi.techfak.scie.descriptors.City;
import de.unibi.techfak.scie.descriptors.Country;
import de.unibi.techfak.scie.descriptors.IsIn;
import de.unibi.techfak.scie.classifiers.ClassifierEvaluation.SetEvaluation;
import de.unibi.techfak.scie.classifiers.annotators.RelationAnnotator;
import de.unibi.techfak.scie.classifiers.annotators.SlotCandidate;
import de.unibi.techfak.scie.classifiers.annotators.SlotSpecification;
import de.unibi.techfak.scie.classifiers.data.ClassificationResult;
import de.unibi.techfak.scie.classifiers.data.CoreDataPoint;
import de.unibi.techfak.scie.classifiers.data.CoreSlotCombinationDataPoint;
import de.unibi.techfak.scie.classifiers.data.FeatureMap;
import de.unibi.techfak.scie.classifiers.data.LabeledDataPoint;
import de.unibi.techfak.scie.classifiers.data.RawRelation;
import de.unibi.techfak.scie.classifiers.data.RelationDataPoint;
import de.unibi.techfak.scie.classifiers.data.TrainingDataReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import junit.framework.TestCase;
import org.apache.uima.analysis_component.JCasAnnotator_ImplBase;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.factory.JCasFactory;
import org.apache.uima.fit.util.CasIOUtil;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

/**
 * This test class is more of an tutorial to enable users to transfer
 * this relation extraction mechanism to their own projects. This
 * tutorial is tied to the README.html shipped with this source code. It
 * extracts "isIn"-relations from an easy sample text.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielfeld.de
 */
public class RelationExtractionTest extends TestCase {

	public RelationExtractionTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	public void testRelationExtraction() throws Exception {
		//first set up our sample text.
		final JCas jcas = JCasFactory.createJCas("src/main/resources/de/unibi/techfak/scie/descriptors/TrainExampleTypeSystem");
		//this is an excerpt of the Wikipedia-Artikel on the Trans-Mongolian Railway
		//http://en.wikipedia.org/wiki/Trans-Mongolian_Railway
		jcas.setDocumentText("The Trans-Mongolian Railway connects Ulan-Ude, "
			+ "on the Trans-Baikal (Trans-Siberian) railway in Russia, "
			+ "with the Chinese city of Jining, by way of Ulaanbaatar in Mongolia.");
		/*
		 * PRE-PROCESSING
		 */
		//tokenize the text (see below for tokenizer implementation)
		final Tokenizer tokenizer = new Tokenizer();
		tokenizer.process(jcas);
		//then apply the city and country annotator (see below for implementation).
		final CityAnnotator cityAnnotator = new CityAnnotator();
		cityAnnotator.process(jcas);
		final CountryAnnotator countryAnnotator = new CountryAnnotator();
		countryAnnotator.process(jcas);
		//that is enough. Now we can save that as an XCAS file.
		CasIOUtil.writeXCas(jcas, new File("trainTestXCAS.xml"));
		/*
		 * CREATING YOUR TRAINING DATA
		 */
		/*
		 * Now that we have a valid XCAS file we can create a .rel file.
		 * Usually you would do that manually. In this case we create it
		 * from program code for clarity.
		 *
		 * In our ID implementation the IDs will be set as follows:
		 *
		 * Ulan-Ude:	1
		 * Jining:	2
		 * Ulaanbaatar:	3
		 * Mongolian:	4
		 * Russia:	5
		 * Chinese:	6
		 * Mongolia:	7
		 *
		 * Therefore our positive relation examples are as follows.
		 */
		final BufferedWriter trainingDataOut = new BufferedWriter(
			new FileWriter("trainData.rel"));
		trainingDataOut.write("ISIN:1,5\n");
		trainingDataOut.write("ISIN:2,6\n");
		trainingDataOut.write("ISIN:3,7\n");
		trainingDataOut.close();
		/*
		 * CLASSIFIER ARCHITECTURE
		 */
		//The implementations can be seen below.
		/*
		 * READING YOUR TRAINING DATA
		 */
		/*
		 * The implementations of our training data readers can be seen below.
		 *
		 * The reading itself then is quite easy. I do it for the example
		 * of our CoreClassifier here.
		 */
		//read the raw relations from the .rel file first.
		final ArrayList<RawRelation> rawRelations
			= RawRelation.readRawRelations(new File("trainData.rel"));
		//then load the XCAS file.
		final JCas readJCAS = JCasFactory.createJCas("src/main/resources/de/unibi/techfak/scie/descriptors/TrainExampleTypeSystem");
		CasIOUtil.readXCas(readJCAS, new File("trainTestXCAS.xml"));
		//then use the TrainingDataReader.
		//initialize an empty training data point collection.
		final ArrayList<LabeledDataPoint> countryTrainingData = new ArrayList<>();
		final TrainingDataReader countryReader = new CountryDataPoint.TrainingDataReader();
		countryReader.readTrainingData(jcas, rawRelations, countryTrainingData);
		//now we are ready to train that classifier.
		/*
		 * TRAINING YOUR CLASSIFIERS
		 */
		/*
		 * The TrainingDataReader automatically creates negative training
		 * examples. In our case we have more positive than negative examples,
		 * but usually it is the other way around: Many more negative than
		 * positive examples. In any way it's smart to balance the data set,
		 * most of the time, by copying random data points of the under-
		 * represented label.
		 */
		TrainingUtils.balanceDataSet(countryTrainingData);
		/*
		 * Now we create our classifier. Here I use a L1-norm-regularized
		 * logistic regression classifier in the LibLinear implementation by
		 * bwaldvogel.
		 *
		 * You can also implement your own classifiers if you implement
		 * the Classifier interface.
		 *
		 * This implicitly uses some defaults for Liblinear. I will not
		 * elaborate on that here.
		 */
		final Classifier countryClassifier = new LibLinearClassifier(SolverType.L1R_LR);
		//now we can directly train ...
		countryClassifier.train(countryTrainingData);
		//and evaluate our result.
		SetEvaluation eval = new SetEvaluation(countryClassifier, countryTrainingData, 10);
		System.out.println("");
		System.out.println("--- Country Classifier Training Results ---");
		System.out.println(eval.toString());
		System.out.println("");
		/*
		 * In this case we do massive overfitting of course. The TrainingUtil-
		 * class provides more sophisticated methods, especially for
		 * cross-validation. For our small example that would be too much.
		 */
		/*
		 * you can also store and load classifiers, which is specified in the
		 * interface. We will not do it here.
		 */
		//Now we train the CityClassifier.
		final ArrayList<LabeledDataPoint> cityTrainingData = new ArrayList<>();
		final TrainingDataReader cityReader = new CityDataPoint.TrainingDataReader();
		cityReader.readTrainingData(jcas, rawRelations, cityTrainingData);
		TrainingUtils.balanceDataSet(cityTrainingData);
		final Classifier cityClassifier = new LibLinearClassifier(SolverType.L1R_LR);
		cityClassifier.train(cityTrainingData);
		eval = new SetEvaluation(cityClassifier, cityTrainingData, 10);
		System.out.println("");
		System.out.println("--- City Classifier Training Results ---");
		System.out.println(eval.toString());
		System.out.println("");
		/*
		 * As you can see: The training process is exactly the same, even though
		 * it are two different kinds of data points used here. This enables
		 * you to implement a more abstract training scheme fitting for your
		 * domain.
		 */
		//To train the relation classifier we need our already trained classifiers.
		//other than that it is the same.
		final ArrayList<LabeledDataPoint> relationTrainingData = new ArrayList<>();
		final TrainingDataReader relationReader = new IsInDataPoint.TrainingDataReader(
			countryClassifier, new Classifier[]{cityClassifier});
		relationReader.readTrainingData(jcas, rawRelations, relationTrainingData);
		TrainingUtils.balanceDataSet(relationTrainingData);
		final Classifier relationClassifier = new LibLinearClassifier(SolverType.L1R_LR);
		relationClassifier.train(relationTrainingData);
		eval = new SetEvaluation(relationClassifier, relationTrainingData, 10);
		System.out.println("");
		System.out.println("--- Relation Classifier Training Results ---");
		System.out.println(eval.toString());
		System.out.println("");
		/*
		 * This is not yet perfect, because our classifier allows the
		 * city slot not to be filled at all. This can be counteracted in the
		 * next step.
		 */
		/*
		 * CREATING RELATION ANNOTATIONS
		 */
		/*
		 * The implementation for the RelationAnnotator can be found below.
		 * Now we can instantiate our RelationAnnotator and apply it.
		 */
		final IsInAnnotator relationAnnotator = new IsInAnnotator(countryClassifier,
			cityClassifier, relationClassifier);
		relationAnnotator.process(jcas);
		/*
		 * Now we are finished and can inspect our result.
		 */
		System.out.println("--- Annotation Results ---");
		for (final IsIn relation : JCasUtil.select(jcas, IsIn.class)) {
			System.out.println("Relation: " + relation.getCoveredText());
			System.out.println("Subject: " + relation.getSubject().getCoveredText());
			System.out.println("Object: " + relation.getObject().getCoveredText());
			System.out.println("");
		}
	}

	/**
	 * In this case we can apply a very simple tokenization: A sentence is just
	 * something that ends with ., ! or ? and a word is something matching on
	 * [\\w-]+
	 */
	private static class Tokenizer extends JCasAnnotator_ImplBase {

		private static final Pattern SENTENCEBOUNDARYPATTERN = Pattern.compile("[!?.]");
		private static final Pattern WORDPATTERN = Pattern.compile("[\\w-]+");

		@Override
		public void process(JCas jcas) throws AnalysisEngineProcessException {
			//first tokenize sentences
			final Matcher sentenceMatcher
				= SENTENCEBOUNDARYPATTERN.matcher(jcas.getDocumentText());
			int begin = 0;
			while (sentenceMatcher.find()) {
				//we found a sentence boundary.
				final Sentence sentence = new Sentence(jcas);
				sentence.setBegin(begin);
				sentence.setEnd(sentenceMatcher.start());
				sentence.addToIndexes();
				begin = sentenceMatcher.end() + 1;
			}
			//then tokenize words.
			final Matcher wordMatcher = WORDPATTERN.matcher(jcas.getDocumentText());
			while (wordMatcher.find()) {
				final Token word = new Token(jcas, wordMatcher.start(), wordMatcher.end());
				word.addToIndexes();
				//also add the single words if a dash is part of the word.
				final String wordText = word.getCoveredText();
				final String[] split = wordText.split("-");
				if (split.length > 1) {
					int idx = wordMatcher.start();
					for (int i = 0; i < split.length; i++) {
						final Token splitWord = new Token(jcas);
						splitWord.setBegin(idx);
						idx += split[i].length();
						splitWord.setEnd(idx);
						splitWord.addToIndexes();
						idx++;
					}
				}
			}
		}
	}

	/**
	 * This is a trivial way to generate globally unique IDs (not thread-safe,
	 * of course). For most purposes this will be complex enough.
	 */
	private static class IDGenerator {

		private static long counter = 0;

		public static long getNewID() {
			return ++counter;
		}
	}

	/**
	 * An Annotator for Cities.
	 */
	private static class CityAnnotator extends JCasAnnotator_ImplBase {

		private static final HashSet<String> cityDatabase = setUpCityDatabase();

		private static HashSet<String> setUpCityDatabase() {
			final HashSet<String> out = new HashSet<>();
			out.add("Ulan-Ude");
			out.add("Jining");
			out.add("Ulaanbaatar");
			return out;
		}

		@Override
		public void process(JCas jcas) throws AnalysisEngineProcessException {
			//iterate over the tokens found and annotate cities.
			/*
			 * This implementation here is, of course, somewhat trivial.
			 * In reality you would be more likely to use a database for
			 * cities which - in most cases - would probably already contain
			 * the relations to the countries there are in. But for the sake
			 * of this tutorial bear with me here.
			 */
			for (final Token word : JCasUtil.select(jcas, Token.class)) {
				if (cityDatabase.contains(word.getCoveredText())) {
					final City city = new City(jcas, word.getBegin(), word.getEnd());
					city.setAnnotationId(IDGenerator.getNewID());
					city.addToIndexes();
				}
			}
		}
	}

	/**
	 * An Annotator for Countries.
	 */
	private static class CountryAnnotator extends JCasAnnotator_ImplBase {

		private static final HashSet<String> countryDatabase = setupCountryDatabase();

		private static HashSet<String> setupCountryDatabase() {
			final HashSet<String> out = new HashSet<>();
			out.add("Russia");
			out.add("China");
			out.add("Chinese");
			out.add("Mongolia");
			out.add("Mongolian");
			return out;
		}

		@Override
		public void process(JCas jcas) throws AnalysisEngineProcessException {
			//iterate over the tokens found and annotate countries.
			/*
			 * This implementation here is, of course, somewhat trivial.
			 * In reality you would be more likely to use a database for
			 * countries.
			 */
			for (final Token word : JCasUtil.select(jcas, Token.class)) {
				if (countryDatabase.contains(word.getCoveredText())) {
					final Country country = new Country(jcas, word.getBegin(), word.getEnd());
					country.setAnnotationId(IDGenerator.getNewID());
					country.addToIndexes();
				}
			}
		}
	}

	/**
	 * Implementation of the CoreDataPoint. In this case we regard a country
	 * to be the trigger. We could also have regarded the City as a trigger.
	 * The implementation is very simple, as you can see.
	 */
	private static class CountryDataPoint extends CoreDataPoint<Country> {

		public CountryDataPoint(Country core, JCas jcas) {
			super(core, jcas);
		}

		@Override
		public void addSpecialFeatures(FeatureMap stdFeatures, boolean addFeaturesToDictionary) {
			//we have no special features here.
		}

		private static class TrainingDataReader extends CoreTrainingDataReader<Country> {

			public TrainingDataReader() {
				//Here we should specify the core class and the
				//name of the relation we used in the .rel file.
				super(Country.class, "ISIN");
			}

			@Override
			public CoreDataPoint<Country> createDataPoint(Country core, JCas jcas) {
				return new CountryDataPoint(core, jcas);
			}
		}
	}

	/**
	 * Implementation of the CoreSlotCombinationDataPoint for the city.
	 * Here the implementation is also pretty self-explanatory. The only
	 * special thing is the "sigma" parameter. Its semantics can be described
	 * as: The average number of words you would expect to lie between a
	 * city and a country in the text. For practical purposes a value between
	 * 6 and 15 is normal.
	 */
	private static class CityDataPoint extends CoreSlotCombinationDataPoint<Country, City> {

		public CityDataPoint(Country core, City slot, JCas jcas) {
			//here we just want to specify the sigma parameter. Otherwise we
			//can redirect everything to the constructor of the super class.
			super(core, slot, jcas, 6);
		}

		@Override
		public void addSpecialFeatures(FeatureMap stdFeatures, boolean addFeaturesToDictionary) {
			//we have no special features here.
		}

		private static class TrainingDataReader extends CoreSlotTrainingDataReader<Country, City> {

			public TrainingDataReader() {
				//Here we should specify the core class, the slot class and the
				//name of the relation we used in the .rel file.
				super(Country.class, City.class, "ISIN");
			}

			@Override
			public CoreSlotCombinationDataPoint<Country, City>
				createCombination(Country core, City slot, JCas jcas) {
				return new CityDataPoint(core, slot, jcas);
			}
		}
	}

	/**
	 * This is the most complex implementation and I will explain it in the
	 * code itself.
	 */
	private static class IsInDataPoint extends RelationDataPoint<Country> {

		public IsInDataPoint(Country core, ClassificationResult coreClassification,
			SlotCandidate[] slots, JCas jcas) {
			super(core, coreClassification, slots, jcas);
		}

		@Override
		public void addSpecialFeatures(FeatureMap stdFeatures, boolean addFeaturesToDictionary) {
			//we have no special features here.
		}

		private static class TrainingDataReader extends RelationTrainingDataReader<Country> {

			public TrainingDataReader(Classifier coreClassifier, Classifier[] slotClassifiers) {
				super(Country.class, coreClassifier, "ISIN");
				/*
				 * Constructing TrainingData for the RelationClassifier depends
				 * on all previous classifications. Therefore we need those as
				 * arguments here. In this case we have only one slot for the
				 * country.
				 *
				 * The Core Classifier can be directly given to the super-class
				 * using its constructor. The slots, however, have to be
				 * specified using the getSlotSpecifications().add() method.
				 */
				if (slotClassifiers.length != 1) {
					throw new UnsupportedOperationException(
						"We expected only one slot classifier!");
				}
				getSlotSpecifications().add(
					new SlotSpecification(City.class, slotClassifiers[0]));
			}

			@Override
			public RelationDataPoint<Country> createDataPoint(
				Country core, ClassificationResult coreClassification,
				SlotCandidate[] slotCandidates, JCas jcas) {
				//here we can just call the constructor
				return new IsInDataPoint(core, coreClassification, slotCandidates, jcas);
			}

			@Override
			public CoreDataPoint<Country> createCoreDataPoint(Country core, JCas jcas) {
				//here too.
				return new CountryDataPoint(core, jcas);
			}

			@Override
			public CoreSlotCombinationDataPoint<Country, ? extends Annotation>
				createCoreSlotCombinationDataPoint(
					SlotSpecification spec, Country core, Annotation slot, JCas jcas) {
						//but here we need some special handling.
					//we must choose the correct constructor ourselves.
					if (slot instanceof City) {
						return new CityDataPoint(core, (City) slot, jcas);
					} else {
						throw new UnsupportedOperationException(
							"Unexpected slot class: " + slot.getClass().getName());
					}
				}
		}
	}

	/**
	 * For this implementation we can mostly copy the implementations
	 * from our TrainingDataReader for the IsInDataPoint
	 *
	 * Only the constructor and the createRelationObject method are interesting.
	 */
	private static class IsInAnnotator extends RelationAnnotator<Country, IsIn> {

		public IsInAnnotator(Classifier coreClassifier,
			Classifier slotClassifier, Classifier relationClassifier) {
			super(Country.class, coreClassifier, relationClassifier);
			getSlotSpecifications().add(
				new SlotSpecification(City.class, slotClassifier));
		}

		@Override
		public RelationDataPoint<Country> createDataPoint(
			Country core, ClassificationResult coreClassification,
			SlotCandidate[] slotCandidates, JCas jcas) {
			//here we can just call the constructor
			return new IsInDataPoint(core, coreClassification, slotCandidates, jcas);
		}

		@Override
		public CoreDataPoint<Country> createCoreDataPoint(Country core, JCas jcas) {
			//here too.
			return new CountryDataPoint(core, jcas);
		}

		@Override
		public CoreSlotCombinationDataPoint<Country, ? extends Annotation>
			createCoreSlotCombinationDataPoint(
				SlotSpecification spec, Country core, Annotation slot, JCas jcas) {
					//but here we need some special handling.
				//we must choose the correct constructor ourselves.
				if (slot instanceof City) {
					return new CityDataPoint(core, (City) slot, jcas);
				} else {
					throw new UnsupportedOperationException(
						"Unexpected slot class: " + slot.getClass().getName());
				}
			}

			/*
			 * This method allows us to specify
			 * sanity checks. If they are not met we return null.
			 * In this case we want to ensure that the city slot is filled.
			 *
			 * If it is filled we create an isIn annotation from it.
			 * We do not need to set begin and end. The super-class will
			 * take care of that. But we need to fill the slots.
			 */
			@Override
			public IsIn createRelationObject(
				Country core, ClassificationResult coreClassification,
				SlotCandidate[] slots, JCas jcas) {
				if (slots[0] == null || slots[0].getAnnotation() == null) {
					return null;
				}
				final City city = (City) slots[0].getAnnotation();
				final IsIn relationInstance = new IsIn(jcas);
				relationInstance.setSubject(city);
				relationInstance.setObject(core);
				return relationInstance;
			}
	}

}
