/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.techfak.scie.classifiers.annotators;

import de.unibi.techfak.scie.descriptors.Aggregated;
import de.unibi.techfak.scie.descriptors.Annotation;
import de.unibi.techfak.scie.classifiers.Classifier;
import de.unibi.techfak.scie.classifiers.data.ClassificationResult;
import de.unibi.techfak.scie.classifiers.data.CoreDataPoint;
import de.unibi.techfak.scie.classifiers.data.CoreSlotCombinationDataPoint;
import de.unibi.techfak.scie.classifiers.data.DataPoint;
import de.unibi.techfak.scie.classifiers.data.FeatureDictionary;
import de.unibi.techfak.scie.classifiers.data.FeatureMap;
import de.unibi.techfak.scie.classifiers.data.LabeledDataPoint;
import de.unibi.techfak.scie.classifiers.data.RelationDataPoint;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import junit.framework.TestCase;
import org.apache.uima.fit.factory.JCasFactory;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

/**
 * This test constructs Relations for a nonsense Relation specification with a
 * basic Annotation as core and another one as slot. The classifiers for core
 * and slot
 * ensure that the core is only accepted if it is the word "Kern" and the slot
 * is only accepted if it is the word "Schlitz".
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class RelationAnnotatorTest extends TestCase {

	public static final String TYPESYSTEM
			= "src/main/resources/de/unibi/techfak/scie/descriptors/SCIEClassifiers";

	public RelationAnnotatorTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of process method, of class RelationAnnotator.
	 */
	public void testProcess() throws Exception {
		//prepare JCas
		final JCas jcas = JCasFactory.createJCas(TYPESYSTEM);
		final String docText = "Dies also ist des Pudels Kern, der nicht durch einen Post-Schlitz passt.";
		jcas.setDocumentText(docText);
		final Pattern wordPattern = Pattern.compile("\\w+");
		final Matcher matcher = wordPattern.matcher(docText);
		while (matcher.find()) {
			final Annotation anno = new Annotation(jcas,
					matcher.start(), matcher.end());
			anno.addToIndexes();
		}
		//use annotator.
		final TestRelationAnnotator instance = new TestRelationAnnotator();
		instance.setR(1);
		instance.process(jcas);
		//check result.
		final Collection<Aggregated> results = JCasUtil.select(jcas, Aggregated.class);
		assertEquals(1, results.size());
		final Aggregated result = results.iterator().next();
		assertEquals("Kern, der nicht durch einen Post-Schlitz", result.getCoveredText());
		assertEquals(1, result.getProbability(), 0);
	}

	public static final class TestRelationAnnotator extends RelationAnnotator<Annotation, Aggregated> {

		public TestRelationAnnotator() {
			super(Annotation.class, new TestCoreSlotClassifier(),
					new TestRelationClassifier());
			final SlotSpecification<Annotation> slotSpec
					= new SlotSpecification(Annotation.class, new TestCoreSlotClassifier());
			getSlotSpecifications().add(slotSpec);
		}

		public RelationDataPoint<Annotation> createDataPoint(Annotation core,
				ClassificationResult coreClassification,
				SlotCandidate[] slotCandidates, JCas jcas) {
			return new TestRelationDataPoint(core, coreClassification,
					slotCandidates, jcas);
		}

		public CoreDataPoint<Annotation> createCoreDataPoint(Annotation core, JCas jcas) {
			return new TestCoreDataPoint(core, jcas);
		}

		@Override
		public CoreSlotCombinationDataPoint<Annotation, ? extends Annotation>
				createCoreSlotCombinationDataPoint(SlotSpecification spec,
						Annotation code, Annotation slot, JCas jcas) {
			return new TestCoreSlotCombination(code, slot, jcas, 0);
		}

		@Override
		public Aggregated createRelationObject(Annotation core,
				ClassificationResult coreClassification,
				SlotCandidate[] slots, JCas jcas) {
			return new Aggregated(jcas);
		}

	}

	public static class TestCoreSlotClassifier implements Classifier {

		private final FeatureDictionary dict = new FeatureDictionary();

		public TestCoreSlotClassifier() {
			this.dict.addFeature("Kern");
			this.dict.addFeature("slotSchlitz");
		}

		@Override
		public void train(ArrayList<LabeledDataPoint> trainingData) {
			throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
		}

		@Override
		public ClassificationResult classify(DataPoint data) {
			final FeatureMap features
					= data.getFeatureRepresentation(dict, false);
			final boolean label;
			if (data instanceof CoreDataPoint) {
				label = features.getValue("Kern") != null;
			} else if (data instanceof CoreSlotCombinationDataPoint) {
				label = features.getValue("slotSchlitz") != null;
			} else {
				throw new UnsupportedOperationException(
						"Unexpected DataPoint class! " + data.getClass().getName());
			}
			return new ClassificationResult(data, label, 1);
		}

		@Override
		public void writeParamters(Writer parameterOut, Writer featuresOut) throws IOException {
			throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
		}

		@Override
		public void readParameters(Reader parameterIn, Reader featuresIn) throws IOException {
			throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
		}

		@Override
		public FeatureDictionary getFeatureDictionary() {
			throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
		}

	}

	public static final class TestRelationClassifier implements Classifier {

		private final FeatureDictionary dict = new FeatureDictionary();

		public TestRelationClassifier() {
			dict.addFeature("coreConfidence");
			dict.addFeature("confidenceSlot0");
		}

		@Override
		public void train(ArrayList<LabeledDataPoint> trainingData) {
			throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
		}

		@Override
		public ClassificationResult classify(DataPoint data) {
			if (data instanceof RelationDataPoint) {
				final FeatureMap features
						= data.getFeatureRepresentation(dict, false);
				final boolean label = features.getValue("coreConfidence") == 1
						&& features.getValue("confidenceSlot0") != null;
				final double confidence;
				if (label) {
					confidence = 1;
				} else {
					confidence = 0;
				}
				return new ClassificationResult(data, label, confidence);
			} else {
				throw new UnsupportedOperationException(
						"Unexpected DataPoint class! " + data.getClass().getName());
			}
		}

		@Override
		public void writeParamters(Writer parameterOut, Writer featuresOut) throws IOException {
			throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
		}

		@Override
		public void readParameters(Reader parameterIn, Reader featuresIn) throws IOException {
			throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
		}

		@Override
		public FeatureDictionary getFeatureDictionary() {
			throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
		}

	}

	public static final class TestCoreDataPoint extends CoreDataPoint<Annotation> {

		public TestCoreDataPoint(Annotation core, JCas jcas) {
			super(core, jcas);
		}

		@Override
		public void addSpecialFeatures(FeatureMap stdFeatures, boolean addFeaturesToDictionary) {
			//no special features.
		}

	}

	public static final class TestCoreSlotCombination extends CoreSlotCombinationDataPoint<Annotation, Annotation> {

		public TestCoreSlotCombination(Annotation core,
				Annotation slot, JCas jcas, double sigma) {
			super(core, slot, jcas, sigma);
		}

		@Override
		public void addSpecialFeatures(FeatureMap stdFeatures, boolean addFeaturesToDictionary) {
			//no special features.
		}

	}

	public static final class TestRelationDataPoint extends RelationDataPoint<Annotation> {

		public TestRelationDataPoint(Annotation core,
				ClassificationResult coreClassification,
				SlotCandidate[] slots, JCas jcas) {
			super(core, coreClassification, slots, jcas);
		}

		@Override
		public void addSpecialFeatures(FeatureMap stdFeatures, boolean addFeaturesToDictionary) {
			//no special features.
		}

	}

}
